// gps_clock.ino
//
// 日時を知る（GPSモジュール使用）
// 
// Written by Nobuki HIRAMINE
// Last modified : 2011/12/03

#include <LiquidCrystal.h>
#include <SoftwareSerial.h>

#define READBUFFERSIZE (256)
#define DELIMITER	(",")

LiquidCrystal g_lcd( 12, 11, 5, 4, 3, 2 );
SoftwareSerial g_gps( 6, 7 );
char g_szReadBuffer[READBUFFERSIZE] = "";
int  g_iIndexChar = 0;

void setup()
{
	g_gps.begin(9600);

	g_lcd.begin( 16, 2 );
	g_lcd.clear();
}

// 閏年の定義
// 西暦年が4で割り切れる年は閏年
// ただし、西暦年が100で割り切れる年は平年
// ただし、西暦年が400で割り切れる年は閏年
int IsLeapYear( int iYear )
{
	if( ( 0 == iYear % 4
	   && 0 != iYear % 100 )
	 || 0 == iYear % 400 )
	{
		return 1;
	}
	return 0;
}

// 日付を１日進める。
void IncrementDay( int& riYear, int& riMonth, int& riDay )
{
	switch( riMonth )
	{
	case 1:
		if( 31 == riDay){ riMonth = 2; riDay = 1; }
		else		    { riDay++; }
		break;
	case 2:
		if( 28 == riDay || 29 == riDay )
		{
			if( IsLeapYear( riYear ) )
			{	// 閏年（2月は29日まである）
				if( 29 == riDay){ riMonth = 3; riDay = 1; }
				else			{ riDay++; }
			}
			else
			{	// 閏年じゃない（2月は28日まで）
				riMonth = 3; riDay = 1;
			}
		}
		else			{ riDay++; }
		break;
	case 3:
		if( 31 == riDay){ riMonth = 4; riDay = 1; }
		else			{ riDay++; }
		break;
	case 4:
		if( 30 == riDay){ riMonth = 5; riDay = 1; }
		else			{ riDay++; }
		break;
	case 5:
		if( 31 == riDay){ riMonth = 6; riDay = 1; }
		else			{ riDay++; }
		break;
	case 6:
		if( 30 == riDay){ riMonth = 7; riDay = 1; }
		else			{ riDay++; }
		break;
	case 7:
		if( 31 == riDay){ riMonth = 8; riDay = 1; }
		else			{ riDay++; }
		break;
	case 8:
		if( 31 == riDay){ riMonth = 9; riDay = 1; }
		else			{ riDay++; }
		break;
	case 9:
		if( 30 == riDay){ riMonth = 10; riDay = 1; }
		else			{ riDay++; }
		break;
	case 10:
		if( 30 == riDay){ riMonth = 11; riDay = 1; }
		else			{ riDay++; }
		break;
	case 11:
		if( 30 == riDay){ riMonth = 12; riDay = 1; }
		else			{ riDay++; }
		break;
	case 12:
		if( 31 == riDay){ riYear++; riMonth = 1; riDay = 1; }
		else			{ riDay++; }
		break;
	}
}

// センテンスの解析。
// $GPRMCの場合、引数変数に、緯度、経度を入れ、戻り値 1 を返す。
// $GPRMC以外の場合、戻り値は 0 を返す。
int AnalyzeLineString( char szLineString[],
					   int& riYear, int& riMonth, int& riDay,
					   int& riHour, int& riMin, int& riSec )
{
	riYear = 0; riMonth = 0; riDay = 0;
	riHour = 0; riMin = 0; riSec = 0;

	// $GPRMC
	if( 0 != strncmp( "$GPRMC", szLineString, 6 ) )
	{
		return 0;
	}
	
	// $GPRMC,085120.307,A,3541.1493,N,13945.3994,E,000.0,240.3,181211,,,A*6A
	strtok( szLineString, DELIMITER );	// $GPRMC
	char* pszUTCTime = strtok( NULL, DELIMITER );	// UTC時刻(hhmmss.ss)
	strtok( NULL, DELIMITER );	// ステータス
	strtok( NULL, DELIMITER );	// 緯度
	strtok( NULL, DELIMITER );	// 北緯か南緯か
	strtok( NULL, DELIMITER );	// 経度
	strtok( NULL, DELIMITER );	// 東経か西経か
	strtok( NULL, DELIMITER );	// 移動の速度
	strtok( NULL, DELIMITER );	// 移動の真方位
	char* pszUTCDate = strtok( NULL, DELIMITER );	// UTC日付(ddmmyy)
	
	if( NULL == pszUTCDate )
	{
		return 0;
	}

	long hhmmss = atol(pszUTCTime);
	riHour = hhmmss / 10000;
	riMin =  (hhmmss / 100) % 100;
	riSec =  hhmmss % 100;
	
	long ddmmyy = atol(pszUTCDate);
	riYear = 2000 + ddmmyy % 100;
	riMonth = (ddmmyy / 100) % 100;
	riDay = ddmmyy / 10000;

	// 日本標準時は、UTC（世界標準時）よりも９時間進んでいる。
	riHour += 9;
	if( 24 <= riHour )
	{	// 日付変更
		riHour - 24;
		IncrementDay( riYear, riMonth, riDay );
	}
	
	return 1;
}

// １行文字列の読み込み
// 0 : 読み取り途中。1 : 読み取り完了。
int ReadLineString(	SoftwareSerial& serial,
					char szReadBuffer[], const int ciReadBufferSize, int& riIndexChar,
					char szLineString[], const int ciLineStringSize )
{
	while( 1 )
	{
		char c = serial.read();
		if( -1 == c )
		{
			break;
		}
		if( '\r' == c  )
		{	// 終端
			szReadBuffer[riIndexChar] = '\0';
			strncpy( szLineString, szReadBuffer, ciLineStringSize - 1 );
			szLineString[ciLineStringSize - 1] = '\0';
			riIndexChar = 0;
			return 1;
		}
		else if( '\n' == c )
		{
			;	// 何もしない
		}
		else
		{	// 途中
			if( (ciReadBufferSize - 1) > riIndexChar )
			{
				szReadBuffer[riIndexChar] = c;
				riIndexChar++;
			}
		}
	}

	return 0;
}

void loop()
{
	char szLineString[READBUFFERSIZE];
	if( !ReadLineString( g_gps,
						 g_szReadBuffer, READBUFFERSIZE, g_iIndexChar,
						 szLineString, READBUFFERSIZE ) )
	{	// 読み取り途中
		return;
	}
	// 読み取り完了

	int iYear, iMonth, iDay;
	int iHour, iMin, iSec;
	if( !AnalyzeLineString( szLineString,
							iYear, iMonth, iDay,
							iHour, iMin, iSec ) )
	{
		return;
	}
	// 日時を読み取れた。

	// LCD出力
	char szBuffer[16];
	sprintf( szBuffer, "%04d / %02d / %02d", iYear, iMonth, iDay );
	g_lcd.setCursor( 0, 0 );
	g_lcd.print( szBuffer );
	sprintf( szBuffer, "  %02d : %02d  %02d", iHour, iMin, iSec );
	g_lcd.setCursor( 0, 1 );
	g_lcd.print( szBuffer );
}
